<?php
namespace Zeedhi\Tools\EntityGenerator;

use Zeedhi\Tools\Nominator;
use Zeedhi\Tools\Holder;
use Zeedhi\Tools\DatabaseReader\DTO;

class MetaData {

    /** @var  Inflector */
    protected $inflector;

    protected $metaData = array();

    protected $inverseRelations;

    public function __construct(Inflector $inflector, $clean, $verbose, $inverseRelations = array()) {
        $this->inflector = $inflector;
        $this->clean = $clean;
        $this->verbose = $verbose;
        $this->inverseRelations = $inverseRelations;
    }

    /**
     * @param $message
     * @param $context
     * @return null
     */
    protected function warning($message, $context = array()) {
        echo "[warning] ", str_replace(array_keys($context), $context, $message), PHP_EOL;
    }

    protected function info($message, $context = array()) {
        if($this->verbose){
            echo "[info] ", str_replace(array_keys($context), $context, $message), PHP_EOL;
        }
    }

    /**
     * @param string      $namespaceName
     * @param Nominator\ZhNamespace $parentNamespace
     *
     * @return Nominator\ZhNamespace
     */
    protected function createNamespace($namespaceName, Nominator\ZhNamespace $parentNamespace = null) {
        $currentNamespace = new Nominator\zhNamespace();
        $currentNamespace->setName($namespaceName);
        if ($parentNamespace) {
            $currentNamespace->setParentNamespace($parentNamespace);
        }

        return $currentNamespace;
    }

    protected function createFullNamespace($namespaceName) {
        $parentNamespace = null;
        foreach(explode("\\", $namespaceName) as $namespacePiece) {
            $parentNamespace = $this->createNamespace($namespacePiece, $parentNamespace);
        }

        return $parentNamespace;
    }

    protected function validateTableHasPk(DTO\Table $table) {
        foreach ($table->getConstraints() as $constraint) {
            if($constraint->getConstraintType() == 'P') return true;
        }
        return false;
    }

    protected function createClass(DTO\Table $table, $classType, Nominator\ZhNamespace $namespace) {
        $tableName = $table->getName();
        $className = $this->inflector->classify($tableName);

        $class = new Nominator\zhClass();
        $class->setName($className);
        $class->setNamespace($namespace);
        $class->setType($classType);

        return $class;
    }

    protected function createClassWithNamespace(DTO\Table $table, Nominator\ZhNamespace $parentNamespace, $classType, $namespaceName) {
        $namespace = $this->createNamespace($namespaceName, $parentNamespace);
        return $this->createClass($table, $classType, $namespace);
    }

    protected function createEntityClass(DTO\Table $table, Nominator\ZhNamespace $parentNamespace) {
        return $this->createClassWithNamespace($table, $parentNamespace, Nominator\zhClass::TYPE_ENTITY, "Entities");
    }

    protected function createRepositoryClass(DTO\Table $table, Nominator\ZhNamespace $parentNamespace) {
        return $this->createClassWithNamespace($table, $parentNamespace, Nominator\zhClass::TYPE_REPOSITORY, "Repositories");
    }

    protected function createListenerClass(DTO\Table $table, Nominator\ZhNamespace $parentNamespace) {
        return $this->createClassWithNamespace($table, $parentNamespace, Nominator\zhClass::TYPE_ENTITY_LISTENER, "Listeners");
    }

    protected function retrieveSubModelNamespace(DTO\Table $table, Nominator\ZhNamespace $parentNamespace) {
        $tableName = $table->getName();
        $subModel = null;

        foreach ($this->data["SUBMODELS"] as $submodel) {
            foreach ($submodel->getTables() as $tablename => $table) {
                if($tablename == $tableName){
                    $subModel = $submodel;
                    break;
                }
            }
        }

        if ($subModel) {
            $subModelNamespaceName = $this->inflector->classify($subModel->getName());
            $subModelNamespace = $this->createNamespace($subModelNamespaceName, $parentNamespace);
        } else {
            $subModelNamespace = $parentNamespace;
        }

        return $subModelNamespace;
    }

    protected function getConsColumns($constraintType, $column) {
        if(isset($this->data["CONSTRAINTS"][$constraintType])){
            foreach ($this->data["CONSTRAINTS"][$constraintType][$column->getTable()->getName()] as $constraint) {
                if($consColumns = $constraint->getConsColumns()){
                    foreach ($consColumns as $consColumn) {
                        if($consColumn->getColumnName() == $column->getName()) return true;
                    }
                }
            }
        }
    }

    protected function columnBelongsToAConstraint($constraintType, $column) {
        return $this->getConsColumns($constraintType, $column);
    }

    protected function isColumnAForeignKey($column) {
        return $this->columnBelongsToAConstraint(DTO\Constraint::CONSTRAINT_TYPE_FK, $column);
    }

    protected function isColumnAPrimaryKey($column) {
        return $this->columnBelongsToAConstraint(DTO\Constraint::CONSTRAINT_TYPE_PK, $column);
    }

    protected function isColumnUnique($column) {
        return $this->columnBelongsToAConstraint(DTO\Constraint::CONSTRAINT_TYPE_UNIQUE, $column);
    }

    protected function createEntityFields(Holder\Entity $entity) {
        /** @var DTO\Column[] $columns */
        $table = $entity->getTable();
        $columns = $table->getColumns();

        foreach ($columns as $columnName => $column) {
            $entityField = new Holder\EntityField();
            $entityField->setColumn($column);
            $entityField->setName($this->inflector->propertify($columnName));
            $entityField->setIsNullable($column->getNullable());
            $entityField->setType(TypeConverter::typeFromMetadata($column->getDataType()));
            $entityField->setDefaultValue(TypeConverter::defaultValueFromMetadata($column->getDataDefault()));
            $entityField->setEntity($entity);
            $entityField->setIsFk($this->isColumnAForeignKey($column) ? "Y" : "N");
            $entityField->setIsId($this->isColumnAPrimaryKey($column) ? "Y" : "N");
            $entityField->setIsUnique($this->isColumnUnique($column) ? "Y" : "N");

            $this->metaData["FIELD"][$table->getName()][$columnName] = $entityField;
        }
    }

    /**
     * @param DTO\Constraint $constraint
     * @return Holder\Entity
     */
    protected function getRelatedEntityByConstraint(DTO\Constraint $constraint, $namespace) {
        $table = null;
        foreach ($this->data["CONSTRAINTS"]["P"] as $tableName => $tbl) {
            foreach ($this->data["TABLES"][$tableName]->getConstraints() as $cons) {
                if($cons->getConstraintName() == $constraint->getReferencedConstraintName()){
                    $table = $this->data["TABLES"][$tableName];
                    break;
                }
            }
        }
        if($table){
            $relatedEntity = $this->createEntity($table, $namespace);
            return $relatedEntity;
        }
        return null;
    }

    protected function createEntityRelations(Holder\Entity $entity, $namespaceName, $constraintType) {
        $table = $entity->getTable();
        $tableName = $table->getName();
        $constraints = $this->data["CONSTRAINTS"][$constraintType][$tableName];

        foreach ($constraints as $constraint) {
            $consColumns = $constraint->getConsColumns();
            if($consColumns){
                foreach ($consColumns as $consColumn) {
                    $columnName = $consColumn->getColumnName();
                    if(isset($this->metaData["FIELD"][$tableName][$columnName])){
                        if($this->metaData["FIELD"][$tableName][$columnName]->getIsId() === "Y") continue;
                        $constraintName = $constraint->getConstraintName();
                        // Remember that inverse relations have the same constraint.
                        $entityRelation = new Holder\EntityRelation();
                        $entityRelation->setConstraint($constraint);
                        // The entity field must not be reset, cause the user can set it to avoid overuse in different relations.
                        $entityField = $this->metaData["FIELD"][$tableName][$columnName];
                        $entityRelation->setEntityField($entityField);
                        $entityRelation->setEntity($entity);
                        $entityRelation->setName($this->inflector->relationify($entityRelation->getEntityField()->getName()));
                        $entityRelation->setIsInversed("N");

                        if($constraintType == "F" && $constraint->getReferencedConstraintName()){
                            $referencedEntity = $this->getRelatedEntityByConstraint($constraint, $namespaceName);
                            if($referencedEntity) {
                                $entityRelation->setRelatedEntity($referencedEntity);
                            }
                            unset($this->metaData["FIELD"][$tableName][$columnName]);
                        }

                        $this->metaData["ENTITYRELATION"][$tableName][$constraintName][$columnName] = $entityRelation;
                    }
                }
            }
        }
    }

    protected function generateInvertedRelation($constraint, $newFieldName, $tableName) {
        $constraintName = $constraint ? $constraint->getConstraintName() : false;
        $fksTableName = $constraint->getTable()->getName();
        $consColumns = $constraint->getConsColumns();
        foreach ($consColumns as $consColumn) {
            $columnName = $consColumn->getColumnName();
            if(isset($this->metaData["ENTITYRELATION"][$fksTableName]) && isset($this->metaData["ENTITY"][$tableName])){
                $oneToManyEntity = $this->metaData["ENTITY"][$tableName];
                $manyToOneEntity = $this->metaData["ENTITY"][$fksTableName];
                $inverseRelation = $this->metaData["ENTITYRELATION"][$fksTableName][$constraintName][$columnName];
                $newEntityField = new Holder\EntityField();
                $newEntityField->setEntity($oneToManyEntity);
                $newEntityField->setName($newFieldName);
                $newEntityField->setIsFk("Y");
                $newEntityField->setIsId("N");
                $newEntityField->setIsUnique("N");
                $newEntityField->setIsNullable("Y");
                $newEntityField->setType(Holder\EntityField::TYPE_OBJECT);

                $newRelation = new Holder\EntityRelation();
                $newRelation->setName($newFieldName);
                $newRelation->setEntity($oneToManyEntity);
                $newRelation->setRelatedEntity($manyToOneEntity);
                $newRelation->setConstraint($constraint);
                $newRelation->setInverseRelation($inverseRelation);
                $newRelation->setIsInversed("Y");
                $newRelation->setEntityField($newEntityField);

                $this->metaData["ENTITYRELATION"][$tableName][$constraintName."_INVERSE"][$columnName] = $newRelation;
            }
        }
    }

    protected function createEntity(DTO\Table $table, $namespaceName = null) {
        if(isset($this->metaData["ENTITY"][$table->getName()])){
            return $this->metaData["ENTITY"][$table->getName()];
        }

        $namespace = $namespaceName ? $this->createFullNamespace($namespaceName) : null;
        $entity = new Holder\Entity();
        $entity->setTable($table);

        $subModelNamespace = $this->retrieveSubModelNamespace($table, $namespace);

        $entity->setClass($this->createEntityClass($table, $subModelNamespace));
        $entity->setRepositoryClass($this->createRepositoryClass($table, $subModelNamespace));
        $entity->setListenerClass($this->createListenerClass($table, $subModelNamespace));
        $entity->setName($entity->getClass()->getName());

        return $entity;
    }

    /**
     * Generates entity related objects.
     *
     * @param mixed[]  $data             The tables, constraints and sequences objects.
     * @param string   $namespaceName    The namespace.
     * @param string[] $tablesToGenerate The tables to generate.
     *
     * @return string[] Entities names.
     */
    public function generateEntitiesForTableNames($data, $namespaceName = null, $tablesToGenerate = array()) {
        $this->data = $data;

        $entitiesNames = array();

        foreach($this->data["TABLES"] as $tableName => $table) {
            if ($this->validateTableHasPk($table)) {
                $entity = $this->createEntity($table, $namespaceName);
                $entityName = $entity->getName();
                $this->createEntityFields($entity);
                $entitiesNames[] = $entityName;
                $this->metaData["ENTITY"][$tableName] = $entity;
                $this->info("Imported entity for table {tableName}", array("tableName" => $tableName));
            } else {
                $this->warning("Table {table} skipped it has no primary key!", array("table"=>$tableName));
            }
        }

        foreach ($this->metaData["ENTITY"] as $entityName=>$entity) {
            /** @var Holder\Entity $entity */
            if(!$this->clean) $this->createEntityRelations($entity, $namespaceName, DTO\Constraint::CONSTRAINT_TYPE_FK);
            $this->createEntityRelations($entity, $namespaceName, DTO\Constraint::CONSTRAINT_TYPE_UNIQUE);
            $this->info("Update entities relations for {entityName}", array("entityName" => $entityName));
        }

        if(!$this->clean){
            foreach($this->inverseRelations as $constraintName => $newFieldName){
                $tableName = null;
                $const = null;
                foreach($this->data["CONSTRAINTS"]["F"] as $constraints){
                    foreach($constraints as $constraint){
                        if($constraint->getConstraintName() == $constraintName){
                            $const = $constraint;
                            break;
                        }
                    }
                }
                if($const){
                    foreach($this->data["CONSTRAINTS"]["P"] as $tblName => $constraints){
                        foreach($constraints as $constraint){
                            if($const->getReferencedConstraintName() == $constraint->getConstraintName()){
                                $tableName = $tblName;
                                break;
                            }
                        }
                    }
                    if($tableName){
                        $this->generateInvertedRelation($const, $newFieldName, $tableName);
                    }
                }
            }
        }
        return $entitiesNames;
    }

    public function getMetaData(){
        return $this->metaData;
    }
}